#ifdef __cplusplus
extern "C" { 
#endif  

#include "f2c.h" 

int sgttrs_(char *trans, integer *n, integer *nrhs, real *dl, real *d__, real *du, real *du2, integer *ipiv, real *b, integer *ldb, integer *info);

#ifdef __cplusplus
}
#endif